from . import _file_exists, _env, get_read_fn, get_write_fn, Timer
import tensorflow as tf
from threading import Thread
import tempfile
import math
import os
if _env['tqdm']:
    from tqdm.auto import tqdm, trange

from .logger import get_logger
logger = get_logger()

'''
dataset_features = {
        'x': {
            'length': 416,
            'names': ["input_ids", "attention_mask"]
        },
        'y': {
            'length': 96,
            'names': ["target_ids", "target_attention_mask", "shifted_target_input_ids"]
        }
    }
'''


def _int64_feature(value):
    return tf.train.Feature(int64_list=tf.train.Int64List(value=value))

def setup_tf_serialization_features(dataset_features):
    global _tf_example_features, _tf_dataset_features
    _tf_dataset_features = dataset_features
    _tf_example_features = []
    for axis in dataset_features:
        _tf_example_features += dataset_features[axis]['names']

def serialize_example(ex):
    feature = {feat: _int64_feature(ex[feat]) for feat in _tf_example_features}
    example_proto = tf.train.Example(features=tf.train.Features(feature=feature))
    return example_proto.SerializeToString()

def SerializeWorker(ex):
    result = serialize_example(ex)
    return result


class TFileIO:
    def __init__(self, outfile=None, tmpfile=None, copy=False, gcopy=False, remove_tmp=True):
        self._outfile, self._rm, = outfile, remove_tmp
        self._tmpfile = tmpfile
        self._copy, self._gcopy = copy, gcopy
        if self._tmpfile:
            self.filename = self._tmpfile
        else:
            self.filename = self._outfile
        self.fn = tf.io.TFRecordWriter(self.filename)
        self.writer = self.fn.write

    def write(self, x):
        self.writer(x)

    def close(self):
        self.fn.flush()
        self.fn.close()
        if (self._gcopy or self._copy) and self._tmpfile:
            _threaded_copy = Thread(target=self._copydaemon,)
            _threaded_copy.start()

    def _copydaemon(self):
        if self._gcopy:
            tf.io.gfile.copy(self._tmpfile, self._outfile, overwrite=True)
            if self._rm:
                if tf.io.gfile.exists(self._outfile):
                    tf.io.gfile.remove(self.filename)
                else:
                    logger.info('Cannot confirm copied file exists. Not removing tmp file')

        elif self._copy:
            with get_write_fn(self._outfile) as write_fn:
                data = tf.io.read_file(self.filename)
                write_fn.write(data)
                write_fn.flush()
            write_fn.close()
            if self._rm:
                tf.io.gfile.remove(self.filename)
        logger.info(f'Copied {self.filename} -> {self._outfile}')

    def __enter__(self):
        return self

    def __exit__(self, *_):
        logger.info(f"Closing File  - {self.filename}")
        self.close()



class TFRWriter(object):
    def __init__(self, dir_path, num_examples, start_idx=1, split='train', write_string='{}_shard_{}.tfrecords', shard_size=50000, overwrite=False, use_tempdir=False):
        self._init_stats()
        self._dir, self._total, self._split, self._fstring,  self._shardsize, self._overwrite = dir_path, num_examples, split, write_string, shard_size, overwrite
        self._startidx = start_idx
        self._usetmp = use_tempdir
        self._shardsize = self._total if self._total < self._shardsize else self._shardsize
        self._numfiles = math.ceil(self._total / self._shardsize)
        self.pbar = trange(self._total, desc=f"[{self._split}] TFRecords Writer", dynamic_ncols=True) if _env['tqdm'] else None
        self._setup_files()
        self.fn = self.openfile()
        self.filename = self.fn.filename
        self.writer = self.fn.write
        self.timer = Timer()
        logger.info(f"- Writer Config for Split [{self._split}] -> Shard Size/Num Ex per Record: {self._shardsize}, Total Items: {self._total}, Write Path: {self._dir}, Number of Files: {self._numfiles}")
        self.timer.start('TFRecords Writer')

    def write(self, x):
        if self.curr_idx >= self._shardsize:
            self.writer(x)
            self.close()
            self.shard += 1
            self.idx += 1
            self.fn = self.openfile()
            self.filename = self.fn.filename
            self.writer = self.fn.write

        else:
            self.writer(x)
            self.idx += 1
            self.curr_idx += 1
        
        if self.pbar:
            self.pbar.update()

    def close(self):
        logger.info(f'Closing. Shard {self.shard}/{self._numfiles}\nFilename: {self.filename}\n{self.idx} Examples written in {self.timer.time()}.\nRemaining Items: {self._total - self.idx}/{self._total}.')        
        self.fn.close()    
    
    def openfile(self):
        self.curr_idx = 0
        filename = self._fstring.format(self._split, self.shard)
        _outfile = os.path.join(self._dir, filename)
        _tmpfile = os.path.join(self._tmpdir, filename) if self._tmpdir else None
        _file = TFileIO(outfile=_outfile, tmpfile=_tmpfile, copy=self._copy, gcopy=self._gcopy, remove_tmp=True)
        if self.pbar:
            self.pbar.set_description(f"[{self._split}] TFRecords Writer: {filename}")
        return _file

    def _init_stats(self):
        self.idx = 0
        self.curr_idx = 0
        self.shard = 1

    def _setup_files(self):
        self._tmpdir = None
        self._existing_fns = None
        self._copy, self._gcopy = False, False
        if self._dir.startswith('s3') or self._usetmp:
            self._tmpdir = tempfile.gettempdir()
            self._copy = True
            if self._usetmp and self._dir.startswith('gs://'):
                self._gcopy = True
        
        if self._dir.startswith('/') or self._dir.startswith('gs://'):
            if self._dir.endswith('/'):
                self._dir = self._dir[:-1]
            if not tf.io.gfile.exists(self._dir):
                tf.io.gfile.makedirs(self._dir)
            _fpath = os.path.join(f'{self._dir}/{self._split}*')
            if tf.io.gfile.exists(self._dir):
                self._existing_fns = tf.io.gfile.glob(_fpath)
                logger.info(f'Existing Files: {self._existing_fns}')
            else:
                logger.info(f'{_fpath} does not exist. Creating Directory.')
                tf.io.gfile.makedirs(self._dir)

        if self._startidx > 1:
            self.shard = self._startidx
            logger.info(f' Manually setting Shard IDX to {self.shard}')
        elif not self._overwrite and self._existing_fns:
            self.shard = len(self._existing_fns) + 1
            logger.info(f' Setting Shard IDX to {self.shard}')
        elif self._overwrite:
            logger.info(f'Overwrite is enabled. Will Start at Shard IDX 1')
        

    def __enter__(self):
        return self

    def __exit__(self, *_):
        logger.info(f"- TFRecords Writer is Closing")
        self.close()


import random
import os
from . import line_count, create_idx_key, get_idx_key, get_write_fn, get_read_fn, _io_type
from . import _env, parser, json, glob, Timer
if _env['tqdm']:
    from tqdm.auto import tqdm, trange
if _env['tf']:
    from .tflow import setup_tf_serialization_features, serialize_example, SerializeWorker, TFRWriter

#if _env['ray']:
#    import ray.util.multiprocessing as mp
#else:
from .logger import get_logger
import multiprocessing as mp
import hashlib
logger = get_logger()

# https://stackoverflow.com/questions/620367/how-to-jump-to-a-particular-line-in-a-huge-text-file
class LineSeekableFile:
    def __init__(self, seekable):
        self.fin = seekable
        self.line_map = list() # Map from line index -> file position.
        self.line_map.append(0)
        while seekable.readline():
            self.line_map.append(seekable.tell())

    def index(self):
        return self.line_map
    
    def __len__(self):
        return len(self.line_map)

    def __getitem__(self, index):
        # NOTE: This assumes that you're not reading the file sequentially.  
        # For that, just use 'for line in file'.
        self.fin.seek(self.line_map[index])
        return self.fin.readline()


class LazyLoadFile:
    def __init__(self, filename, skip_broken=True):
        self.filename = filename
        self.reader = get_read_fn(filename)
        self._skip = skip_broken
        self.fseek = LineSeekableFile(self.reader)
        if self._skip:
            self.lineidx = 0
            self.badidx = 0
    
    def random_iter(self, num_lines=None):
        num_lines = num_lines if num_lines else len(self.fseek)
        total_idx = [i for i in range(num_lines)]
        random.shuffle(total_idx)
        for idx in total_idx:
            if self._skip:
                try:
                    yield self.loads(self.fseek[idx])
                    self.lineidx += 1
                except:
                    self.badidx += 1

            else:
                yield self.loads(self.fseek[idx])
    
    def quick_iter(self, num_lines=None):
        num_lines = num_lines if num_lines else len(self.fseek)
        for x, line in enumerate(self.reader):
            if self._skip:
                try:
                    yield self.loads(line)
                    self.lineidx += 1
                except:
                    self.badidx += 1
            else:
                yield self.loads(line)
                if x >= num_lines:
                    break

    def iter(self):
        for line in self.reader:
            if self._skip:
                try:
                    yield self.loads(line)
                    self.lineidx += 1
                except:
                    self.badidx += 1
            else:
                yield self.loads(line)

    def loads(self, v):
        return parser.parse(v).as_dict()

    def __getitem__(self, idx):
        return self.loads(self.fseek[idx])

    def __len__(self):
        return len(self.fseek)
    
    def stats(self):
        return {'loaded': self.lineidx, 'missed': self.badidx}
    
    def resetstats(self):
        self.lineidx = 0
        self.badidx = 0
        return {'loaded': self.lineidx, 'missed': self.badidx}


def setup_tokenize_fn(tokenizer_fn):
    assert _env['transformers'], 'Transformers must be installed to use tokenize function'
    global _tokenize_fn
    _tokenize_fn = tokenizer_fn

def TokenizerWorker(ex):
    result = _tokenize_fn(ex)
    return result

def setup_iter_fn(iter_fn):
    global _iter_func
    _iter_func = iter_fn

def IterWorker(ex):
    result = _iter_func(ex)
    return result

def setup_filter_fns(filter_fns):
    global _filter_func
    _filter_func = filter_fns

def FilterWorker(ex):
    result = {}
    for key in ex:
        if key not in _filter_func['bypass'] and key in _filter_func:
            res = _filter_func[key](ex[key])
            if res:
                result[key] = res
        elif key in _filter_func['bypass']:
            result[key] = ex[key]
    if bool(result):
        return None
    return result


def FileIterator(filename):
    with get_read_fn(filename) as f:
        for line in f:
            yield parser.parse(line).as_dict()
    raise StopIteration


def make_hashes(inputs):
    return hashlib.sha256(str.encode(inputs)).hexdigest()

def check_hashes(inputs, hashed_text):
	if make_hashes(inputs) == hashed_text:
		return hashed_text
	return False

class Pylines:
    def __init__(self, input_fns=None, output_fn=None, skip_broken=True, overwrite_output=False, use_lazy=False, use_mp=True, use_idx=False, total_lines=0):
        self._skip, self._lazy, self._mp, self._idx, self._overwrite = skip_broken, use_lazy, use_mp, use_idx, overwrite_output
        self.total_lines = total_lines
        self.writer, self.reader = None, None
        self.input_fns, self.output_fn = None, None
        self.stats = {}
        self.timer = Timer()
        self._io(input_fns, output_fn)
    
    def as_tokenizer(self, tokenizer_fn=None, input_fns=None, use_mp=True):
        if tokenizer_fn:
            setup_tokenize_fn(tokenizer_fn)
        assert _tokenize_fn, 'tokenizer_fn must first be set before being able to run'
        self._io(input_fns, output_fn=None)
        for result in self._as_iter(_tokenize_fn, TokenizerWorker, use_mp, desc='Tokenization'):
            yield result
        logger.info(f'{self.timer.stop()} for Tokenizing {self.total_lines} Items')

    def run_tokenizer(self, tokenizer_fn=None, input_fns=None, output_fn=None, use_mp=True):
        self._io(input_fns, output_fn)
        for result in self.as_tokenizer(tokenizer_fn=tokenizer_fn, use_mp=use_mp):
            self.write(result)
        self.flush()

    def as_processor(self, iter_func=None, input_fns=None, use_mp=True):
        if iter_func:
            setup_iter_fn(iter_func)
        assert _iter_func, 'iter_func must first be set before running'
        self._io(input_fns, output_fn=None)
        for result in self._as_iter(_iter_func, IterWorker, use_mp, desc='Iterator Function'):
            yield result
        logger.info(f'{self.timer.stop()} for {self.total_lines} Items')

    def run_processor(self, iter_func=None, input_fns=None, output_fn=None, use_mp=True):
        self._io(input_fns, output_fn)
        for result in self.as_processor(iter_func=iter_func, use_mp=use_mp):
            self.write(result)
        self.flush()

    # filter_funcs = {'text': filter_fuc, 'target': filter_func, 'idx': filter_func, 'bypass': ['key_1', 'key_2']}
    def as_filter(self, filter_funcs=None, input_fns=None, use_mp=True):
        if filter_funcs:
            setup_filter_fns(filter_funcs)
        assert _filter_func, 'filter_funcs must first be set before running'
        self._io(input_fns, output_fn=None)
        for result in self._as_iter(FilterWorker, FilterWorker, use_mp, desc='Filtering Items'):
            yield result
        logger.info(f'{self.timer.stop()} for Filtering {self.total_lines} Items')
    
    def run_filter(self, filter_funcs=None, input_fns=None, output_fn=None, use_mp=True):
        self._io(input_fns, output_fn)
        for result in self.as_filter(filter_funcs=filter_funcs, use_mp=use_mp):
            self.write(result)
        self.flush()
    
    def as_encoder(self, dataset_features=None, tokenizer_fn=None, serialization='tf', input_fns=None, use_mp=True):
        _methods = ['tf']
        assert serialization in _methods, f'Currently only {_methods} are supported'
        assert _env['tf'], 'Tensorflow is required to run Serialization'
        if dataset_features:
            for axis in dataset_features:
                assert 'names' in dataset_features[axis], 'names is a required key for dataset features.'
            setup_tf_serialization_features(dataset_features)

        self._io(input_fns, output_fn=None)
        all_results = list()
        if _tokenize_fn or tokenizer_fn:
            for result in self.as_tokenizer(tokenizer_fn, use_mp=use_mp):
                all_results.append(result)
        else:
            logger.warning(f'No Tokenizer Function is Found. Assuming Input Files are Pretokenized.')
            for result in self.as_iterator():
                all_results.append(result)
        
        for serialized_ex in self._as_iter_items(all_results, serialize_example, SerializeWorker, use_mp=use_mp, desc=f'Serializing to {serialization}'):
            yield serialized_ex
        logger.info(f'{self.timer.stop()} for Serializing {len(all_results)} Examples')

    def run_encoder(self, output_dir, dataset_features=None, tokenizer_fn=None, serialization='tf', input_fns=None, start_idx=1, split_key='split', split='train', write_string='{}_shard_{}.tfrecords', shard_size=50000, overwrite=False, use_tempdir=False, use_mp=True):
        self._io(input_fns, output_fn=None)
        _total_match = self.count_matching(split_key, split) if split_key else self.total_lines
        with TFRWriter(output_dir, _total_match, start_idx, split, write_string, shard_size, overwrite, use_tempdir) as writer:
            for serialized_ex in self.as_encoder(dataset_features, tokenizer_fn, use_mp=use_mp):
                writer.write(serialized_ex)
        
        #writer.close()


    def _as_iter(self, IterFunc, Worker, use_mp, desc):
        pbar = trange(self.total_lines, desc=desc) if _env['tqdm'] else None
        self.timer.start(desc)
        if use_mp:
            if isinstance(use_mp, int):
                pool = mp.Pool(use_mp)
            else:
                pool = mp.Pool()
            for fn in self.input_fns:
                for result in pool.imap_unordered(Worker, FileIterator(fn)):
                    yield result
                    if pbar:
                        pbar.update()
            
        else:
            for fn in self.input_fns:
                for result in self._file_iter(fn):
                    ex = IterFunc(result)
                    yield ex
                    if pbar:
                        pbar.update()
        if pbar:
            pbar.close()

    def _as_iter_items(self, items, IterFunc, Worker, use_mp, desc):
        pbar = trange(len(items), desc=desc) if _env['tqdm'] else None
        self.timer.start(desc)
        if use_mp:
            if isinstance(use_mp, int):
                pool = mp.Pool(use_mp)
            else:
                pool = mp.Pool()
            for result in pool.imap_unordered(Worker, items):
                yield result
                if pbar:
                    pbar.update()
            
        else:
            for item in items:
                ex = IterFunc(item)
                yield ex
                if pbar:
                    pbar.update()
        if pbar:
            pbar.close()

    def deduplicate(self, keys, input_fns=None, output_fn=None, write=True):
        self._io(input_fns, output_fn)
        _sets = {}
        results = list()
        assert _io_type(keys) == 'list', 'Keys must be in the form of a list'
        for key in keys:
            _sets[key] = set()
        for result in self.as_iterator():
            _pass = True
            for k in keys:
                hashed_key = make_hashes(result[k])
                if hashed_key in _sets[k]:
                    _pass = False
                else:
                    _sets[k].add(hashed_key)
            if _pass:
                if write:
                    self.write(result)
                else:
                    results.append(result)
        
        if not write:
            return results

    def find(self, key, value, results='first', filename=None):
        assert results in ['first', 'all'], 'Results should either be all or first to return'
        _matched = False
        if filename:
            for x, result in enumerate(self._fast_file_iter(filename)):
                if result[key] == value:
                    yield result.as_dict()
                    _matched = True
                
                if _matched:
                    print(f'Found Match on IDX: {x}')
                    if results == 'first':
                        break
                    elif results == 'all':
                        _matched = False

                
        else:
            for fn in self.input_fns:
                for x, result in enumerate(self._fast_file_iter(fn)):
                    if result[key] == value:
                        yield result.as_dict()
                        _matched = True
                        if results == 'first':
                            break
                
                    if _matched:
                        print(f'Found Match on IDX: {x} in {fn}')
                        if results == 'first':
                            break
                        elif results == 'all':
                            _matched = False

                if _matched and results == 'first':
                    break
    
    def merge(self, input_fns=None, output_fn=None):
        self._io(input_fns, output_fn)
        pbar = trange(self.total_lines, desc=f'Merging {len(self.input_fns)} Files') if _env['tqdm'] else None
        self.timer.start(f'Merging {len(self.input_fns)} Files')
        for result in self.as_iterator():
            self.write(result)
            if pbar:
                pbar.update()
        self.flush()
        if pbar:
            pbar.close()
        logger.info(f'{self.timer.stop()} with {self.total_lines} Items to {self.output_fn}')
        
    def count_matching(self, key, value, input_fns=None):
        _matches = 0
        self._io(input_fns)
        for result in self.as_iterator():
            if result[key] == value:
                _matches += 1
        
        return _matches

    def verify_linecount(self):
        _idx = 0
        for result in self.as_iterator():
            if result and result != '':
                _idx += 1

        logger.info(f'Original Count: {self.total_lines} Verified Count: {_idx}')
        if _idx >= self.total_lines:
            self.total_lines = _idx
        return _idx

    def write(self, item):
        if not self.writer:
            assert self.output_fn, 'Output File must be set prior to write. call .writefile(filename) to set the output file'
            self.writer_fn = get_write_fn(self.output_fn, overwrite=self._overwrite)
            self.writer = self.writer_fn.write
        
        self.writer(json.dumps(item, ensure_ascii=False))
        self.writer('\n')

    def index(self, idx, fn=None):
        if fn:
            return self._file_idx(idx, fn)
        else:
            results = {}
            for fn in self.input_fns:
                results[fn] = self._file_idx(idx, fn)
            return results
    
    def to_dict(self, input_fns=None):
        results = {}
        self._io(input_fns)
        for result in self.as_iterator():
            for key in result:
                if key not in results:
                    results[key] = {'items': [result[key]], 'count': 1}
                else:
                    results[key]['items'].append(result[key])
                    results[key]['count'] += 1
        
        return results

    def to_list(self, input_fns=None):
        results = list()
        self._io(input_fns)
        for result in self.as_iterator():
            results.append(result)
        return results

    def from_list(self, all_items, output_fn=None):
        self._io(input_fns=None, output_fn=output_fn)
        assert _io_type(all_items) == 'list', 'This function must be used with a list'
        for item in all_items:
            self.write(item)

    def to_batches(self, all_items, batch_size=2):
        all_batches = list()
        for batch in self._build_batches(all_items, batch_size):
            all_batches += [batch]
        return all_batches

    def as_iterator(self):
        for fn in self.input_fns:
            for item in self._file_iter(fn):
                yield item


    def _build_batches(self, lst, batch_size):
        for i in range(0, len(lst), batch_size):
            yield lst[i:i + batch_size]

    def _file_idx(self, idx, fn):
        reader = get_read_fn(fn)
        for x, line in enumerate(reader):
            if x < idx:
                pass
            elif x == idx:
                return self.loads(line)
        reader.close()

    def _file_iter(self, fn):
        reader = get_read_fn(fn)
        if self._skip:
            self._reset_stats(fn)
        for line in reader:
            if self._skip:
                try:
                    yield self.loads(line)
                    self.stats[fn]['read'] += 1
                except:
                    self.stats[fn]['missed'] += 1
            else:
                yield self.loads(line)
        reader.close()

    def _fast_file_iter(self, fn):
        reader = get_read_fn(fn)
        if self._skip:
            self._reset_stats(fn)
        for line in reader:
            if self._skip:
                try:
                    yield self.loads(line)
                    self.stats[fn]['read'] += 1
                except:
                    self.stats[fn]['missed'] += 1

            else:
                yield self.loads(line)
        reader.close()

    def _get_file_lines(self, input_fns=None):
        if input_fns:
            for fn in input_fns:
                self.total_lines += line_count(fn)
        else:
            self.total_lines = 0
            for fn in self.input_fns:
                self.total_lines += line_count(fn)

    def __iter__(self):
        for fn in self.input_fns:
            for result in self._file_iter(fn):
                yield result
    
    def _io(self, input_fns=None, output_fn=None):
        if input_fns:
            self._setup_input_fns(input_fns)
            self._get_file_lines()
        if output_fn:
            self._setup_output_fn(output_fn)

    def _setup_input_fns(self, input_fns):
        in_files = []
        if _io_type(input_fns) == 'str':
            if input_fns.endswith('*'):
                in_files = glob(input_fns)
            else:
                in_files = [input_fns]
        elif _io_type(input_fns) == 'list':
            for fn in input_fns:
                if fn.endswith('*'):
                    in_files += glob(fn)
                else:
                    in_files.append(fn)
        else:
            raise ValueError('Input Filenames should be a string or list')
        
        if self.input_fns:
            in_files = [f for f in in_files if f not in self.input_fns]
            if len(in_files) != 0:
                self.input_fns += in_files
        else:
            self.input_fns = in_files
    
    def _setup_output_fn(self, output_fn):
        if _io_type(output_fn) == 'str':
            self.output_fn = output_fn
        else:
            raise ValueError('Output Filenames should be a string')


    def parse(self, v):
        return parser.parse(v)

    def loads(self, v):
        return parser.parse(v).as_dict()

    def load(self, v):
        return json.load(v)
    
    def dumps(self, v):
        return json.dumps(v, ensure_ascii=False)

    def dump(self, fn, v):
        if _io_type(fn) == 'str':
            json.dump(v, get_write_fn(fn, overwrite=self._overwrite))
        else:
            json.dump(v, fn)
    
    def _reset_stats(self, fn):
        self.stats[fn] = {'read': 0, 'missed': 0}

    def clear_input_files(self):
        self.stats = {}
        self.input_fns = None

    def set_input_files(self, input_fns):
        self.clear_input_files()
        self._setup_input_fns(input_fns)

    def add_files(self, input_fns):
        self._setup_input_fns(input_fns)

    def set_writefile(self, output_fn, overwrite=False):
        self._overwrite = overwrite
        self._setup_output_fn(output_fn)

    def close(self):
        if self.writer:
            self.writer_fn.close()
        if self.reader:
            self.reader.close()

    def flush(self):
        if self.writer:
            self.writer_fn.flush()

    def linecount(self, filename=None):
        results = {}
        if filename:
            results[filename] = line_count(filename)
            return results
        for fn in self.input_fns:
            results[fn] = line_count(fn)
        return results


    def __len__(self):
        return self.total_lines        

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.close()


import os
import os.path
from setuptools import setup, find_packages

root = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(root, 'README.md'), 'rb') as readme:
    long_description = readme.read().decode('utf-8')

setup(
    name='pylines',
    package_dir={'': 'src'},
    packages=find_packages('src'),
    version='0.0.3',
    description='work with large jsonline files with ease',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Tri Songz',
    author_email='ts@scontentengine.ai',
    url='http://github.com/trisongz/pylines',
    keywords=['json', 'json lines', 'jsonlines'],
    zip_safe=False,
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
    ],
    python_requires='>3.4',
    install_requires=[
        "pysimdjson"
    ],
    extras_require={
        'cloud': [
            'tensorflow>=2.1.0',
            'google-cloud-storage',
            'smart_open[all]'
        ],
        'nlp': [
            'transformers',
        ],
        'all': [
            'tensorflow>=2.1.0',
            'torch>=1.0',
            'google-cloud-storage',
            'smart_open[all]',
            'transformers',
        ]
    }
)


__version__ = "0.0.3"

try:
    import absl.logging
except ImportError:
    pass
else:
    absl.logging.set_verbosity("info")
    absl.logging.set_stderrthreshold("info")
    absl.logging._warn_preinit_stderr = False

import os
import subprocess
import simdjson as json
import time
import tempfile

parser = json.Parser()

try:
    import tensorflow as tf
    glob = tf.io.gfile.glob
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
    _tf_available = True
except ImportError:
    import glob
    glob = glob.glob
    _tf_available = False

try:
    import redis
    _redis_available = True
except ImportError:
    _redis_available = False

try:
    import transformers
    _transformers_available = True
except ImportError:
    _transformers_available = False

try:
    import smart_open
    _smartopen_available = True
except ImportError:
    _smartopen_available = False

try:
    import ray
    _ray_available = True
except ImportError:
    _ray_available = False

try:
    import torch
    _torch_available = True
except ImportError:
    _torch_available = False

try:
    import tqdm.auto
    _tqdm_available = True
except ImportError:
    _tqdm_available = False

try:
    from google.cloud import storage
    from google.oauth2 import service_account
    USE_GCS = os.environ.get("GOOGLE_APPLICATION_CREDENTIALS", None)
    if USE_GCS:
        if os.path.exists(USE_GCS):
            gcp_client = service_account.Credentials.service_account_info = json.load(open(USE_GCS, 'r'))
            gcp_storage_client = storage.Client.from_service_account_json(USE_GCS)
            _gcs_available = True
        else:
            gcp_client, gcp_storage_client = None, None
            _gcs_available = False
except ImportError:
    gcp_client, gcp_storage_client = None, None
    _gcs_available = False

try:
    import boto3
    _S3_ID = os.environ.get("AWS_ACCESS_KEY_ID", None)
    _S3_KEY = os.environ.get("AWS_SECRET_ACCESS_KEY", None)
    _BOTO_PATH_1 = os.environ.get("AWS_SHARED_CREDENTIALS_FILE", None)
    _BOTO_PATH_2 = os.path.join(os.environ.get("HOME", "/root"), ".aws/credentials")
    _BOTO_PATH_2_EXISTS = os.path.exists(_BOTO_PATH_2)
    _BOTO_PATH_3 = os.environ.get("BOTO_CONFIG", None)

    _BOTO_PATH_4 = os.environ.get("BOTO_PATH", None)
    _BOTO_PATH_5 = os.path.join(os.environ.get("HOME", "/root"), ".boto")
    _BOTO_PATH_5_EXISTS = os.path.exists(_BOTO_PATH_5)
    s3_client = None

    if _S3_ID and _S3_KEY:
        s3_client = boto3.Session(aws_access_key_id=_S3_ID, aws_secret_access_key=_S3_KEY)
    elif _BOTO_PATH_1 or _BOTO_PATH_2_EXISTS or _BOTO_PATH_3:
        s3_client = boto3.Session()
    elif _BOTO_PATH_4:
        os.environ['BOTO_CONFIG'] = _BOTO_PATH_4
        s3_client = boto3.Session()
    elif _BOTO_PATH_5_EXISTS:
        os.environ['BOTO_CONFIG'] = _BOTO_PATH_5
        s3_client = boto3.Session()
    
    if s3_client:
        _s3_available = True
    else:
        _s3_available = False
except ImportError:
    _s3_available = False
    s3_client = None



_env = {
    'gcs': _gcs_available,
    's3': _s3_available,
    'transformers': _transformers_available,
    'smartopen': _smartopen_available,
    'torch': _torch_available,
    'tf': _tf_available,
    'ray': _ray_available,
    'tqdm': _tqdm_available,
}


def auth_cloud(gcs=None, s3=None):
    # expects a json file for gcs
    global _gcs_available, _s3_available, s3_client, gcp_client, gcp_storage_client
    if gcs:
        try:
            gcp_client = service_account.Credentials.service_account_info = json.load(open(gcs, 'r'))
            gcp_storage_client = storage.Client.from_service_account_json(gcs)
            _gcs_available = True
        except Exception as e:
            print(f'Unable to Authenticate GCS: {str(e)}')
            _gcs_available = False

    if s3:
        if _io_type(s3) == 'str' and os.path.exists(s3):
            os.environ['AWS_SHARED_CREDENTIALS_FILE'] = s3
            try:
                s3_client = boto3.Session()
                _s3_available = True
            except Exception as e:
                print(f'Unable to Authenticate S3: {str(e)}')
                _s3_available = False
        elif _io_type(s3) == 'dict':
            if 'aws_access_key_id' in s3 and 'aws_secret_access_key' in s3:
                try:
                    s3_client = boto3.Session(aws_access_key_id=s3['aws_access_key_id'], aws_secret_access_key=s3['aws_secret_access_key'])
                    _s3_available = True
                except Exception as e:
                    print(f'Unable to Authenticate S3: {str(e)}')
                    _s3_available = False
            elif 'aws_access_key_id'.upper() in s3 and 'aws_secret_access_key'.upper() in s3:
                try:
                    s3_client = boto3.Session(aws_access_key_id=s3['aws_access_key_id'.upper()], aws_secret_access_key=s3['aws_secret_access_key'.upper()])
                    _s3_available = True
                except Exception as e:
                    print(f'Unable to Authenticate S3: {str(e)}')
                    _s3_available = False
            elif 'AWS_SHARED_CREDENTIALS_FILE' in s3:
                os.environ['AWS_SHARED_CREDENTIALS_FILE'] = s3
                try:
                    s3_client = boto3.Session()
                    _s3_available = True
                except Exception as e:
                    print(f'Unable to Authenticate S3: {str(e)}')
                    _s3_available = False
    else:
        _S3_ID = os.environ.get("AWS_ACCESS_KEY_ID", None)
        _S3_KEY = os.environ.get("AWS_SECRET_ACCESS_KEY", None)
        _S3_PATH = os.environ.get("AWS_SHARED_CREDENTIALS_FILE", None)
        _GCS_PATH = os.environ.get("GOOGLE_APPLICATION_CREDENTIALS", None)
        if not _s3_available:
            if _S3_ID and _S3_KEY:
                try:
                    s3_client = boto3.Session(aws_access_key_id=_S3_ID, aws_secret_access_key=_S3_KEY)
                    _s3_available = True
                except Exception as e:
                    print(f'Unable to Authenticate S3: {str(e)}')
                    _s3_available = False
            elif _S3_PATH:
                try:
                    os.environ['AWS_SHARED_CREDENTIALS_FILE'] = _S3_PATH
                    s3_client = boto3.Session()
                    _s3_available = True
                except Exception as e:
                    print(f'Unable to Authenticate S3: {str(e)}')
                    _s3_available = False
        if not _gcs_available and _GCS_PATH:
            try:
                gcp_client = service_account.Credentials.service_account_info = json.load(open(_GCS_PATH, 'r'))
                gcp_storage_client = storage.Client.from_service_account_json(_GCS_PATH)
                _gcs_available = True
            except Exception as e:
                print(f'Unable to Authenticate GCS: {str(e)}')
                _gcs_available = False

    if _gcs_available:
        print('Authenticated GCS')
    else:
        gcp_client = None
        gcp_storage_client = None

    if _s3_available:
        print('Authenticated S3')
    else:
        s3_client = None

def get_read_fn(filename):
    if filename.startswith('gs://'):
        if _tf_available:
            return tf.io.gfile.GFile(filename, 'rb+')
        elif _smartopen_available:
            if _gcs_available:
                return smart_open.open(filename, 'rb', transport_params=dict(client=gcp_storage_client))
            else:
                return smart_open.open(filename, 'rb')
        else:
            raise ValueError('Tensorflow and SmartOpen are not available to open a GCS File')
    elif filename.startswith('s3://'):
        if _smartopen_available:
            if _s3_available:
                return smart_open.open(filename, 'rb', transport_params={'session': s3_client})
            else:
                return smart_open.open(filename, 'rb')
        else:
            raise ValueError('SmartOpen is not available to open a S3 File')
    elif filename.startswith('https://'):
        if _smartopen_available:
            return smart_open.open(filename, 'rb')
        else:
            raise ValueError('SmartOpen is not available to open a HTTP File')
    elif filename.startswith('hdfs://') or filename.startswith('webhdfs://'):
        if _smartopen_available:
            return smart_open.open(filename, 'rb')
        else:
            raise ValueError('SmartOpen is not available to open a HDFS File')
    else:
        if _tf_available:
            return tf.io.gfile.GFile(filename, 'rb+')
        elif _smartopen_available:
            return smart_open.open(filename, 'rb')
        else:
            return open(filename, 'rb')

def _file_exists(filename):
    if _tf_available:
        return tf.io.gfile.exists(filename)
    else:
        return os.path.exists(filename)

def get_write_fn(filename, overwrite=False):
    f_exists = _file_exists(filename)
    if overwrite:
        _write_mode = 'wb'
    elif f_exists:
        _write_mode = 'ab'
    else:
        _write_mode = 'wb'
    if _tf_available:
        _write_mode = _write_mode + '+'
    if filename.startswith('gs://'):
        if _tf_available:
            return tf.io.gfile.GFile(filename, _write_mode)
        elif _smartopen_available:
            if _gcs_available:
                return smart_open.open(filename, _write_mode, transport_params=dict(client=gcp_storage_client))
            else:
                return smart_open.open(filename, _write_mode)
        else:
            raise ValueError('Tensorflow and SmartOpen are not available to open a GCS File')
    elif filename.startswith('s3://'):
        if _smartopen_available:
            if _s3_available:
                return smart_open.open(filename, _write_mode, transport_params={'session': s3_client})
            else:
                return smart_open.open(filename, _write_mode)
        else:
            raise ValueError('SmartOpen is not available to open a S3 File')
    elif filename.startswith('https://'):
        if _smartopen_available:
            return smart_open.open(filename, _write_mode)
        else:
            raise ValueError('SmartOpen is not available to open a HTTP File')
    elif filename.startswith('hdfs://') or filename.startswith('webhdfs://'):
        if _smartopen_available:
            return smart_open.open(filename, _write_mode)
        else:
            raise ValueError('SmartOpen is not available to open a HDFS File')
    else:
        if _tf_available:
            return tf.io.gfile.GFile(filename, _write_mode)
        elif _smartopen_available:
            return smart_open.open(filename, _write_mode)
        else:
            return open(filename, _write_mode)


def line_count(filename):
    return int(subprocess.check_output(['wc', '-l', filename]).split()[0])

def create_idx_key(filename, line):
    idx = {i: key for i, key, v in enumerate(line.items())}
    idx_fn = (filename.split('.')[-1].strip() + '.idx')
    json.dump(idx, get_write_fn(idx_fn), overwrite=True)
    return idx

def get_idx_key(filename):
    idx_fn = (filename.split('.')[-1].strip() + '.idx')
    if _file_exists(idx_fn):
        idx_key = json.loads(get_read_fn(idx_fn))
        idx = {v: key for key, v in idx_key.items()}
        return idx
    else:
        return None


def _io_type(x):
    if isinstance(x, list):
        return 'list'
    elif isinstance(x, dict):
        return 'dict'
    elif isinstance(x, bool):
        return 'bool'
    elif isinstance(x, str):
        return 'str'
    elif isinstance(x, int):
        return 'int'
    elif isinstance(x, float):
        return 'float'

class Timer:
    def __init__(self):
        self.idx = 0
        self.active = False

    def start(self, task):
        if self.active:
            self.stop()

        self.task = task
        self.start_timer()
        self.active = True

    def start_timer(self):
        self.start_time = time.time()
        self.ckpt_time = time.time()
        self.total_time = 0
    
    def time(self):
        stop_time = time.time()
        time_string = f'- Current Time for {self.task}: {(stop_time - self.start_time) / 60:.2f} mins / {(stop_time - self.start_time) :.2f} secs'
        return time_string
    
    def secs(self):
        stop_time = time.time()
        s = (stop_time - self.start_time)
        return s
    
    def stop(self):
        stop_time = time.time()
        time_string = f'- Total Time for {self.task}: {(stop_time - self.start_time) / 60:.2f} mins / {(stop_time - self.start_time) :.2f} secs'
        self.start_time = 0
        return time_string


from . import logger
from . import io
from .io import Pylines, LazyLoadFile, LineSeekableFile


# Imports

import threading
import os
import sys
import logging
from typing import Optional

_lock = threading.Lock()
_pylines_handler: Optional[logging.Handler] = None

class LogFormatter(logging.Formatter):

    COLOR_CODES = {
        logging.CRITICAL: "\033[38;5;196m", # bright/bold magenta
        logging.ERROR:    "\033[38;5;9m", # bright/bold red
        logging.WARNING:  "\033[38;5;11m", # bright/bold yellow
        logging.INFO:     "\033[38;5;111m", # white / light gray
        logging.DEBUG:    "\033[1;30m"  # bright/bold black / dark gray
    }

    RESET_CODE = "\033[0m"
    def __init__(self, color, *args, **kwargs):
        super(LogFormatter, self).__init__(*args, **kwargs)
        self.color = color

    def format(self, record, *args, **kwargs):
        if (self.color == True and record.levelno in self.COLOR_CODES):
            record.color_on  = self.COLOR_CODES[record.levelno]
            record.color_off = self.RESET_CODE
        else:
            record.color_on  = ""
            record.color_off = ""
        return super(LogFormatter, self).format(record, *args, **kwargs)

class PylinesLogger:
    def __init__(self, config):
        self.config = config
        self.logger = self.setup_logging()
    
    def setup_logging(self):
        logger = logging.getLogger()
        logger.setLevel(logging.DEBUG)
        if self.config["console_log_output"] == "stdout":
            console_log_output = sys.stdout
        else:
            console_log_output = sys.stderr
        
        console_handler = logging.StreamHandler(console_log_output)
        console_handler.setLevel(self.config["console_log_level"].upper())
        console_formatter = LogFormatter(fmt=self.config["log_line_template"], color=self.config["console_log_color"])
        console_handler.setFormatter(console_formatter)
        logger.addHandler(console_handler)
        return logger

    def get_logger(self):
        return self.logger

def _setup_library_root_logger(name):
    logger_config = {
        'console_log_output': "stdout", 
        'console_log_level': "info",
        'console_log_color': True,
        'logfile_file': None,
        'logfile_log_level': "debug",
        'logfile_log_color': False,
        'log_line_template': f"%(color_on)s[{name}] %(funcName)-5s%(color_off)s: %(message)s"
    }
    logger = PylinesLogger(logger_config)
    return logger.get_logger()


def _configure_library_root_logger(name="Pylines") -> None:
    global _pylines_handler
    with _lock:
        if _pylines_handler:
            return
        _pylines_handler = _setup_library_root_logger(name)
        _pylines_handler.propagate = False


def get_logger(name: Optional[str] = "Pylines") -> logging.Logger:
    """
    Return a logger with the specified name.
    This function is not supposed to be directly accessed unless you are writing a custom transformers module.
    """
    if name is None:
        name = "Pylines"
    _configure_library_root_logger(name)
    return _pylines_handler